import os
import tempfile
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from fast_forward.ranking import Ranking

RUN = {
    "q1": {"d0": 1, "d1": 2, "d2": 300},
    "q2": {"d0": 4, "d1": 5, "d2": 600, "d3": 7},
}
DUMMY_QUERIES = {"q1": "query 1", "q2": "query 2"}


class TestRanking(unittest.TestCase):
    def setUp(self):
        self.ranking = Ranking.from_run(RUN)
        self.ranking_with_queries = Ranking.from_run(RUN, queries=DUMMY_QUERIES)

    def test_properties(self):
        self.assertEqual({"q1", "q2"}, self.ranking.q_ids)
        self.assertEqual(len(self.ranking), 2)
        self.assertIn("q1", self.ranking)
        self.assertIn("q2", self.ranking)
        self.assertNotIn("q3", self.ranking)

    def test_attach_queries(self):
        self.assertFalse(self.ranking.has_queries)
        self.assertTrue(self.ranking_with_queries.has_queries)

        # check whether queries are attached to the correct query IDs
        self.assertEqual(
            pd.unique(
                self.ranking_with_queries._df.loc[
                    self.ranking._df["q_id"].eq("q1"), "query"
                ]
            ).tolist(),
            ["query 1"],
        )
        self.assertEqual(
            pd.unique(
                self.ranking_with_queries._df.loc[
                    self.ranking._df["q_id"].eq("q2"), "query"
                ]
            ).tolist(),
            ["query 2"],
        )

        # test whether error is raised when queries are incomplete
        more_queries = {"qx": "other query"}
        with self.assertRaises(ValueError):
            Ranking.from_run(RUN, queries=more_queries)

        more_queries.update(DUMMY_QUERIES)
        r2_with_queries = Ranking.from_run(RUN, queries=more_queries)
        self.assertEqual(r2_with_queries, self.ranking_with_queries)

    def test_eq(self):
        r1 = Ranking.from_run({"q1": {"d1": 1, "d2": 2}})
        r2 = Ranking.from_run({"q1": {"d2": 2, "d1": 1}})
        r3 = Ranking.from_run({"q1": {"d1": 2, "d2": 3}})
        r4 = Ranking.from_run({"q1": {"d1": 1, "d2": 2}, "q2": {}})
        self.assertEqual(r1, r2)
        self.assertNotEqual(r1, r3)
        self.assertEqual(r1, r4)

        self.assertNotEqual(r1, "string")
        self.assertNotEqual(r1, 0)

    def test_operators(self):
        self.assertEqual(self.ranking + 0, self.ranking)
        self.assertEqual(self.ranking * 1, self.ranking)
        self.assertEqual(
            self.ranking + 1,
            Ranking.from_run(
                {
                    "q1": {"d0": 2, "d1": 3, "d2": 301},
                    "q2": {"d0": 5, "d1": 6, "d2": 601, "d3": 8},
                }
            ),
        )
        self.assertEqual(
            self.ranking * 2,
            Ranking.from_run(
                {
                    "q1": {"d0": 2, "d1": 4, "d2": 600},
                    "q2": {"d0": 8, "d1": 10, "d2": 1200, "d3": 14},
                }
            ),
        )
        self.assertEqual(1 + self.ranking, self.ranking + 1)
        self.assertEqual(2 * self.ranking, self.ranking * 2)
        self.assertEqual(self.ranking * 2, self.ranking + self.ranking)

        self.assertTrue((self.ranking_with_queries + 1).has_queries)
        self.assertTrue((self.ranking_with_queries * 2).has_queries)
        self.assertTrue((self.ranking_with_queries + self.ranking).has_queries)
        self.assertTrue((self.ranking + self.ranking_with_queries).has_queries)

        # test whether missing scores are correctly interpreted as 0
        self.assertEqual(
            self.ranking
            + Ranking.from_run({"q1": {"d0": 1, "d3": 1}, "q3": {"d0": 1}}),
            Ranking.from_run(
                {
                    "q1": {"d0": 2, "d1": 2, "d2": 300, "d3": 1},
                    "q2": {"d0": 4, "d1": 5, "d2": 600, "d3": 7},
                    "q3": {"d0": 1},
                }
            ),
        )

    def test_cut(self):
        self.assertEqual(
            self.ranking.cut(2),
            Ranking.from_run({"q1": {"d2": 300, "d1": 2}, "q2": {"d2": 600, "d3": 7}}),
        )
        self.assertTrue(self.ranking_with_queries.cut(2).has_queries)

    def test_save_load(self):
        self.ranking.name = "Dummy"
        fd, f = tempfile.mkstemp()
        f = Path(f)
        self.ranking.save(f)
        r_from_file = Ranking.from_file(f)
        self.assertEqual(self.ranking, r_from_file)
        self.assertEqual(self.ranking.name, r_from_file.name)
        os.close(fd)
        os.remove(f)

    def test_normalize(self):
        self.assertEqual(
            Ranking.from_run(
                {
                    "q1": {"d0": 1, "d1": 2, "d2": 3},
                    "q2": {"d0": 4, "d1": 5, "d2": 6},
                }
            ).normalize(),
            Ranking.from_run(
                {
                    "q1": {"d0": 0, "d1": 1 / 5, "d2": 2 / 5},
                    "q2": {"d0": 3 / 5, "d1": 4 / 5, "d2": 1},
                }
            ),
        )

        self.assertEqual(
            Ranking.from_run({"q1": {"d0": 5, "d1": 5}}).normalize(),
            Ranking.from_run({"q1": {"d0": 0, "d1": 0}}),
        )

        self.assertTrue(self.ranking_with_queries.normalize().has_queries)

    def test_interpolate(self):
        df = self.ranking_with_queries._df.copy()
        df["score"] = list(map(np.float32, range(len(self.ranking._df))))
        r2 = Ranking(df)
        r_int = self.ranking.interpolate(r2, 0.5)
        self.assertNotEqual(self.ranking, r_int)
        self.assertEqual(r_int["q1"], {"d2": 152.0, "d1": 3.5, "d0": 3.5})
        self.assertEqual(r_int["q2"], {"d2": 300.0, "d3": 4.0, "d1": 3.5, "d0": 3.5})
        self.assertTrue(r_int.has_queries)

        r3 = Ranking.from_run({"q1": {"d1": 1, "d2": 2}})
        self.assertEqual(
            r3.interpolate(r3, 0.5, normalize=True),
            Ranking.from_run({"q1": {"d1": 0, "d2": 1}}),
        )

        # test whether missing scores are correctly interpreted as 0
        r4 = Ranking.from_run({"q1": {"d1": 1, "d2": 1}, "q2": {"d0": 1}})
        r5 = Ranking.from_run({"q1": {"d0": 1, "d1": 1}, "q3": {"d0": 1}})
        self.assertEqual(
            r4.interpolate(r5, 0.5),
            Ranking.from_run(
                {
                    "q1": {"d0": 0.5, "d1": 1, "d2": 0.5},
                    "q2": {"d0": 0.5},
                    "q3": {"d0": 0.5},
                }
            ),
        )

        # test whether interpolate yields the same results as manually adding the rankings
        self.assertEqual(r4.interpolate(r5, 0.5), 0.5 * r4 + 0.5 * r5)

    def test_rr_scores(self):
        self.assertEqual(
            self.ranking.rr_scores(k=1),
            Ranking.from_run(
                {
                    "q1": {"d0": 1 / 4, "d1": 1 / 3, "d2": 1 / 2},
                    "q2": {"d0": 1 / 5, "d1": 1 / 4, "d2": 1 / 2, "d3": 1 / 3},
                }
            ),
        )
        self.assertTrue(self.ranking_with_queries.rr_scores().has_queries)


if __name__ == "__main__":
    unittest.main()



""".. include:: ../docs/indexer.md
.. include:: ../docs/util.md
"""  # noqa: D205, D400, D415

from typing import TYPE_CHECKING

import numpy as np
from tqdm import tqdm

from fast_forward.util.indexer import Indexer, IndexingDict

if TYPE_CHECKING:
    from collections.abc import Callable

    import pandas as pd

    from fast_forward.index.base import Index
    from fast_forward.ranking import Ranking

__all__ = [
    "Indexer",
    "IndexingDict",
    "to_ir_measures",
    "cos_dist",
    "create_coalesced_index",
]


def to_ir_measures(ranking: "Ranking") -> "pd.DataFrame":
    """Return a ranking as a data frame suitable for the ir-measures library.

    :param ranking: The input ranking.
    :return: The data frame.
    """
    return ranking._df[["q_id", "id", "score"]].rename(
        columns={"q_id": "query_id", "id": "doc_id"}
    )


def cos_dist(a: np.ndarray, b: np.ndarray) -> float:
    """Cosine distance of two vectors.

    :param a: First vector.
    :param b: Second vector.
    :return: Cosine distance.
    """
    assert len(a.shape) == len(b.shape) == 1
    return float(1 - np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b)))


def create_coalesced_index(
    source_index: "Index",
    target_index: "Index",
    delta: float,
    distance_function: "Callable[[np.ndarray, np.ndarray], float]" = cos_dist,
    batch_size: int | None = None,
) -> None:
    """Create a compressed index using sequential coalescing.

    :param source_index: Source index (containing multiple vectors for each document).
    :param target_index: Target index (must be empty).
    :param delta: The coalescing threshold.
    :param distance_function: The distance function.
    :param batch_size: Use batches instead of adding all vectors at the end.
    :raises ValueError: When the target index is not empty.
    """
    if len(target_index) > 0:
        raise ValueError("Target index is not empty.")

    def _coalesce(P: np.ndarray) -> list[np.ndarray]:
        P_new = []
        A = []
        A_avg = np.empty(())
        first_iteration = True
        for v in P:
            if first_iteration:
                first_iteration = False
            elif distance_function(v, A_avg) >= delta:
                P_new.append(A_avg)
                A = []
            A.append(v)
            A_avg = np.mean(A, axis=0)
        P_new.append(A_avg)
        return P_new

    batch_size = batch_size or len(source_index.doc_ids)
    vectors, doc_ids = [], []
    for doc_id in tqdm(source_index.doc_ids):
        # check if batch is full
        if len(vectors) == batch_size:
            target_index.add(np.array(vectors), doc_ids=doc_ids)
            vectors, doc_ids = [], []

        v_old, _ = source_index._get_vectors([doc_id])
        v_new = _coalesce(v_old)
        vectors.extend(v_new)
        doc_ids.extend([doc_id] * len(v_new))
    if len(vectors) > 0:
        target_index.add(np.array(vectors), doc_ids=doc_ids)

    assert source_index.doc_ids == target_index.doc_ids



from typing import TYPE_CHECKING, Any

import torch
from transformers import AutoModel, AutoTokenizer

from fast_forward.encoder.base import Encoder

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from pathlib import Path

    import numpy as np
    from transformers import BatchEncoding
    from transformers.modeling_outputs import BaseModelOutput


class TransformerEncoder(Encoder):
    """Uses a pre-trained Transformer model for encoding.

    The outputs corresponding to the CLS token from the last hidden layer are used.
    """

    def __init__(
        self,
        model: "str | Path",
        device: str = "cpu",
        model_args: "Mapping[str, Any]" = {},
        tokenizer_args: "Mapping[str, Any]" = {},
        tokenizer_call_args: "Mapping[str, Any]" = {
            "padding": True,
            "truncation": True,
        },
        normalize: bool = False,
    ) -> None:
        """Create a Transformer encoder.

        :param model: Pre-trained Transformer model (name or path).
        :param device: PyTorch device.
        :param model_args: Additional arguments for the model.
        :param tokenizer_args: Additional arguments for the tokenizer.
        :param tokenizer_call_args: Additional arguments for the tokenizer call.
        :param normalize: L2-normalize output representations.
        """
        super().__init__()
        self._model = AutoModel.from_pretrained(model, **model_args)
        self._model.to(device)
        self._model.eval()
        self._tokenizer = AutoTokenizer.from_pretrained(model, **tokenizer_args)
        self._device = device
        self._tokenizer_call_args = tokenizer_call_args
        self._normalize = normalize

    def _get_tokenizer_inputs(self, texts: "Sequence[str]") -> list[str]:
        """Prepare input texts for tokenization.

        :param texts: The texts to encode.
        :return: The tokenizer inputs.
        """
        return list(texts)

    def _aggregate_model_outputs(
        self,
        model_outputs: "BaseModelOutput",
        model_inputs: "BatchEncoding",  # noqa: ARG002
    ) -> torch.Tensor:
        """Aggregate Transformer outputs using the CLS token (last hidden state).

        Encoders overriding this function may make use of `model_inputs`.

        :param model_outputs: The Transformer outputs.
        :param model_inputs: The Transformer inputs (unused).
        :return: The CLS token representations from the last hidden state.
        """
        return model_outputs.last_hidden_state[:, 0]

    def _encode(self, texts: "Sequence[str]") -> "np.ndarray":
        model_inputs = self._tokenizer(
            self._get_tokenizer_inputs(texts),
            return_tensors="pt",
            **self._tokenizer_call_args,
        ).to(self._device)

        with torch.no_grad():
            model_outputs = self._model(**model_inputs)
            result = self._aggregate_model_outputs(model_outputs, model_inputs)
            if self._normalize:
                result = torch.nn.functional.normalize(result, p=2, dim=1)
        return result.cpu().detach().numpy()


class TCTColBERTQueryEncoder(TransformerEncoder):
    """Pre-trained TCT-ColBERT query encoder.

    Adapted from Pyserini:
    https://github.com/castorini/pyserini/blob/310c828211bb3b9528cfd59695184c80825684a2/pyserini/encode/_tct_colbert.py#L72

    Corresponding paper: https://aclanthology.org/2021.repl4nlp-1.17/
    """

    def __init__(
        self,
        model: "str | Path" = "castorini/tct_colbert-msmarco",
        device: str = "cpu",
        max_length: int = 36,
    ) -> None:
        """Create a TCT-ColBERT query encoder.

        :param model: Pre-trained TCT-ColBERT model (name or path).
        :param device: PyTorch device.
        :param max_length: Maximum number of tokens per query.
        """
        self._max_length = max_length
        super().__init__(
            model,
            device=device,
            tokenizer_call_args={
                "max_length": max_length,
                "truncation": True,
                "add_special_tokens": False,
            },
        )

    def _get_tokenizer_inputs(self, texts: "Sequence[str]") -> list[str]:
        return ["[CLS] [Q] " + q + "[MASK]" * self._max_length for q in texts]

    def _aggregate_model_outputs(
        self,
        model_outputs: "BaseModelOutput",
        model_inputs: "BatchEncoding",  # noqa: ARG002
    ) -> torch.Tensor:
        embeddings = model_outputs.last_hidden_state[:, 4:, :]
        return torch.mean(embeddings, dim=-2)


class TCTColBERTDocumentEncoder(TransformerEncoder):
    """Pre-trained TCT-ColBERT document encoder.

    Adapted from Pyserini:
    https://github.com/castorini/pyserini/blob/310c828211bb3b9528cfd59695184c80825684a2/pyserini/encode/_tct_colbert.py#L27

    Corresponding paper: https://aclanthology.org/2021.repl4nlp-1.17/
    """

    def __init__(
        self,
        model: "str | Path" = "castorini/tct_colbert-msmarco",
        device: str = "cpu",
        max_length: int = 512,
    ) -> None:
        """Create a TCT-ColBERT document encoder.

        :param model: Pre-trained TCT-ColBERT model (name or path).
        :param device: PyTorch device.
        :param max_length: Maximum number of tokens per document.
        """
        self._max_length = max_length
        super().__init__(
            model,
            device=device,
            tokenizer_call_args={
                "max_length": max_length,
                "padding": True,
                "truncation": True,
                "add_special_tokens": False,
            },
        )

    def _get_tokenizer_inputs(self, texts: "Sequence[str]") -> list[str]:
        return ["[CLS] [D] " + d for d in texts]

    def _aggregate_model_outputs(
        self,
        model_outputs: "BaseModelOutput",
        model_inputs: "BatchEncoding",
    ) -> torch.Tensor:
        token_embeddings = model_outputs.last_hidden_state[:, 4:, :]
        input_mask_expanded = (
            model_inputs.attention_mask[:, 4:]
            .unsqueeze(-1)
            .expand(token_embeddings.size())
            .float()
        )
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)
        sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
        return sum_embeddings / sum_mask


class TASBEncoder(TransformerEncoder):
    """Pre-trained TAS-B (topic-aware sampling) encoder.

    Corresponding paper: https://dl.acm.org/doi/10.1145/3404835.3462891
    """

    def __init__(
        self,
        model: "str | Path" = "sebastian-hofstaetter/distilbert-dot-tas_b-b256-msmarco",
        device: str = "cpu",
    ) -> None:
        """Create a TAS-B encoder.

        :param model: Pre-trained TAS-B model (name or path).
        :param device: PyTorch device.
        """
        # TAS-B uses CLS-pooling (TransformerEncoder default)
        super().__init__(model, device=device)


class ContrieverEncoder(TransformerEncoder):
    """Pre-trained Contriever encoder.

    Adapted from: https://huggingface.co/facebook/contriever

    Corresponding paper: https://openreview.net/forum?id=jKN1pXi7b0
    """

    def __init__(
        self,
        model: "str | Path" = "facebook/contriever",
        device: str = "cpu",
    ) -> None:
        """Create a Contriever encoder.

        :param model: Pre-trained Contriever model (name or path).
        :param device: PyTorch device.
        """
        super().__init__(model, device=device)

    def _aggregate_model_outputs(
        self,
        model_outputs: "BaseModelOutput",
        model_inputs: "BatchEncoding",
    ) -> torch.Tensor:
        token_embeddings = model_outputs[0].masked_fill(
            ~model_inputs.attention_mask[..., None].bool(), 0.0
        )
        return (
            token_embeddings.sum(dim=1)
            / model_inputs.attention_mask.sum(dim=1)[..., None]
        )


class BGEEncoder(TransformerEncoder):
    """Pre-trained BGE encoder.

    Corresponding paper: https://dl.acm.org/doi/10.1145/3626772.3657878
    """

    def __init__(
        self,
        model: "str | Path" = "BAAI/bge-base-en-v1.5",
        device: str = "cpu",
    ) -> None:
        """Create a BGE encoder.

        :param model: Pre-trained BGE model (name or path).
        :param device: PyTorch device.
        """
        super().__init__(model, device=device, normalize=True)



""".. include:: docs/main.md
.. include:: docs/ranking.md
"""  # noqa: D205, D400, D415

import importlib.metadata

from fast_forward import encoder, index, quantizer, util
from fast_forward.ranking import Ranking

__all__ = ["encoder", "index", "quantizer", "util", "Ranking"]
__version__ = importlib.metadata.version("fast-forward-indexes")



import abc
import logging
from collections.abc import Iterable, Iterator, Sequence
from enum import Enum
from operator import itemgetter
from time import perf_counter

import numpy as np
import pandas as pd
from tqdm import tqdm

from fast_forward.encoder.base import Encoder
from fast_forward.quantizer import Quantizer
from fast_forward.ranking import Ranking

LOGGER = logging.getLogger(__name__)


class Mode(Enum):
    """Enum used to set the ranking mode of an index."""

    PASSAGE = 1
    MAXP = 2
    FIRSTP = 3
    AVEP = 4


IDSequence = Sequence[str | None]


class Index(abc.ABC):
    """Abstract base class for Fast-Forward indexes."""

    _query_encoder: Encoder | None = None
    _quantizer: Quantizer | None = None

    def __init__(
        self,
        query_encoder: Encoder | None = None,
        quantizer: Quantizer | None = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
    ) -> None:
        """Create an index.

        :param query_encoder: The query encoder to use.
        :param quantizer: The quantizer to use.
        :param mode: The ranking mode.
        :param encoder_batch_size: The encoder batch size.
        """
        super().__init__()
        if query_encoder is not None:
            self.query_encoder = query_encoder
        self.mode = mode
        if quantizer is not None:
            self.quantizer = quantizer
        self._encoder_batch_size = encoder_batch_size

    def encode_queries(self, queries: Sequence[str]) -> np.ndarray:
        """Encode queries.

        :param queries: The queries to encode.
        :raises RuntimeError: When no query encoder exists.
        :return: The query representations.
        """
        if self.query_encoder is None:
            raise RuntimeError("Index does not have a query encoder.")

        result = []
        for i in range(0, len(queries), self._encoder_batch_size):
            batch = queries[i : i + self._encoder_batch_size]
            result.append(self.query_encoder(batch))
        return np.concatenate(result)

    @property
    def query_encoder(self) -> Encoder | None:
        """Return the query encoder if it exists.

        :return: The query encoder (if any).
        """
        return self._query_encoder

    @query_encoder.setter
    def query_encoder(self, encoder: Encoder) -> None:
        """Set the query encoder.

        :param encoder: The query encoder to set.
        """
        assert isinstance(encoder, Encoder)
        self._query_encoder = encoder

    @property
    def quantizer(self) -> Quantizer | None:
        """Return the quantizer if it exists.

        :return: The quantizer (if any).
        """
        return self._quantizer

    def _on_quantizer_set(self) -> None:
        """Handle a quantizer being attached to this index."""
        pass

    @quantizer.setter
    def quantizer(self, quantizer: Quantizer) -> None:
        """Set the quantizer.

        This is only possible before any vectors are added to the index.

        :param quantizer: The new quantizer.
        :raises RuntimeError: When the index is not empty.
        """
        assert isinstance(quantizer, Quantizer)

        if len(self) > 0:
            raise RuntimeError("Quantizers can only be attached to empty indexes.")
        self._quantizer = quantizer
        self._on_quantizer_set()
        quantizer.set_attached()

    @property
    def mode(self) -> Mode:
        """Return the ranking mode.

        :return: The ranking mode.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: Mode) -> None:
        """Set the ranking mode.

        :param mode: The ranking mode to set.
        """
        assert isinstance(mode, Mode)
        self._mode = mode

    @abc.abstractmethod
    def _get_internal_dim(self) -> int | None:
        pass

    @property
    def dim(self) -> int | None:
        """Return the dimensionality of the vector index.

        If no vectors exist, return `None`.
        If a quantizer is used, return the dimension of the codes.

        :return: The dimensionality (if any).
        """
        if self._quantizer is not None:
            return self._quantizer.dims[0]
        return self._get_internal_dim()

    @abc.abstractmethod
    def _get_doc_ids(self) -> set[str]:
        pass

    @property
    def doc_ids(self) -> set[str]:
        """Return all unique document IDs.

        :return: The document IDs.
        """
        return self._get_doc_ids()

    @abc.abstractmethod
    def _get_psg_ids(self) -> set[str]:
        pass

    @property
    def psg_ids(self) -> set[str]:
        """Return all unique passage IDs.

        :return: The passage IDs.
        """
        return self._get_psg_ids()

    @abc.abstractmethod
    def _get_num_vectors(self) -> int:
        pass

    def __len__(self) -> int:
        """Return the index size.

        :return: The number of vectors in the index.
        """
        return self._get_num_vectors()

    @abc.abstractmethod
    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Document IDs may have duplicates, passage IDs are assumed to be unique.
        Vectors may be quantized.

        Specific to index implementation.

        :param vectors:
            The representations, shape `(num_vectors, dim)` or
            `(num_vectors, quantized_dim)`.
        :param doc_ids: The corresponding document IDs.
        :param psg_ids: The corresponding passage IDs.
        """
        pass

    def add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence | None = None,
        psg_ids: IDSequence | None = None,
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Only one of `doc_ids` and `psg_ids` may be `None`.
        Individual IDs in the sequence may also be `None`, but each vector must have at
        least one associated ID.
        Document IDs may have duplicates, passage IDs must be unique.

        :param vectors: The representations, shape `(num_vectors, dim)`.
        :param doc_ids: The corresponding document IDs (may be duplicate).
        :param psg_ids: The corresponding passage IDs (must be unique).
        :raises ValueError: When the number of IDs does not match the number of vectors.
        :raises ValueError:
            When the input vector and index dimensionalities do not match.
        :raises ValueError: When a vector has neither a document nor a passage ID.
        :raises RuntimeError: When items can't be added to the index for any reason.
        """
        num_vectors, dim = vectors.shape

        if doc_ids is None:
            doc_ids = [None] * num_vectors
        if psg_ids is None:
            psg_ids = [None] * num_vectors
        if not len(doc_ids) == len(psg_ids) == num_vectors:
            raise ValueError("Number of IDs does not match number of vectors.")

        if self.dim is not None and dim != self.dim:
            raise ValueError(
                f"Input vector dimensionality ({dim}) does not match "
                f"index dimensionality ({self.dim})."
            )

        for doc_id, psg_id in zip(doc_ids, psg_ids):
            if doc_id is None and psg_id is None:
                raise ValueError("Vector has neither document nor passage ID.")

        self._add(
            vectors if self.quantizer is None else self.quantizer.encode(vectors),
            doc_ids,
            psg_ids,
        )

    @abc.abstractmethod
    def _get_vectors(self, ids: Iterable[str]) -> tuple[np.ndarray, list[list[int]]]:
        """Get vectors and corresponding IDs from the index.

        Return a tuple containing:
            * A single array of all vectors necessary to compute the scores for each
                document/passage.
            * For each document/passage (in the same order as the IDs), a list of
                integers (depending on the mode).

        The integers will be used to get the corresponding vectors from the array.
        The output of this function depends on the current mode.
        If a quantizer is used, this function returns quantized vectors.

        Specific to index implementation.

        :param ids: The document/passage IDs to get the representations for.
        :return: The vectors and corresponding indices.
        """
        pass

    def _compute_scores(
        self, data: pd.DataFrame | pd.Series, query_vectors: np.ndarray
    ) -> pd.DataFrame:
        """Compute scores for a data frame.

        The input data frame needs a "q_no" column with unique query numbers.

        :param data: Input data frame (or series).
        :param query_vectors: All query vectors indexed by "q_no".
        :return: Data frame with computed scores.
        """
        # map doc/passage IDs to unique numbers (0 to n)
        id_df = data[["id"]].drop_duplicates().reset_index(drop=True)
        id_df["id_no"] = id_df.index

        # attach doc/passage numbers to data frame
        df = data.merge(id_df, on="id", suffixes=(None, "_")).reset_index(drop=True)

        # get all required vectors from the FF index
        vectors, id_to_vec_idxs = self._get_vectors(id_df["id"].to_list())
        if self.quantizer is not None:
            vectors = self.quantizer.decode(vectors)

        # compute indices for query vectors and doc/passage vectors in current arrays
        select_query_vectors = []
        select_vectors = []
        select_scores = []
        c = 0
        for id_no, q_no in zip(df["id_no"], df["q_no"]):
            vec_idxs = id_to_vec_idxs[id_no]
            select_vectors.extend(vec_idxs)
            select_scores.append(list(range(c, c + len(vec_idxs))))
            c += len(vec_idxs)
            select_query_vectors.extend([q_no] * len(vec_idxs))

        # compute all dot products (scores)
        q_reps = query_vectors[select_query_vectors]
        d_reps = vectors[select_vectors]
        scores = np.sum(q_reps * d_reps, axis=1)

        # select aggregation operation based on current mode
        if self.mode == Mode.MAXP:
            op = np.max
        elif self.mode == Mode.AVEP:
            op = np.average
        else:
            op = itemgetter(0)

        def _mapfunc(i: int) -> float:
            scores_i = select_scores[i]
            if len(scores_i) == 0:
                return np.nan
            return op(scores[scores_i])

        # insert FF scores in the correct rows
        df["ff_score"] = df.index.map(_mapfunc)
        return df

    def _early_stopping(
        self,
        df: pd.DataFrame,
        query_vectors: np.ndarray,
        cutoff: int,
        alpha: float,
        depths: Iterable[int],
    ) -> pd.DataFrame:
        """Compute scores with early stopping for a data frame.

        The input data frame needs a "q_no" column with unique query numbers.

        :param df: Input data frame.
        :param query_vectors: All query vectors indexed by "q_no".
        :param cutoff: Cut-off depth for early stopping.
        :param alpha: Interpolation parameter.
        :param depths: Depths to compute scores at.
        :return: Data frame with computed scores.
        """
        # data frame for computed scores
        scores_so_far = pd.DataFrame()

        # [a, b] is the interval for which the scores are computed in each step
        a = 0
        for b in sorted(depths):
            if b < cutoff:
                continue

            # identify queries which do not meet the early stopping criterion
            if a == 0:
                # first iteration: take all queries
                q_ids_left = pd.unique(df["q_id"])
            else:
                # subsequent iterations: compute ES criterion
                q_ids_left = (
                    scores_so_far.groupby("q_id")
                    .filter(
                        lambda g: g["int_score"].nlargest(cutoff).iat[-1]
                        < alpha * g["score"].iat[-1] + (1 - alpha) * g["ff_score"].max()
                    )["q_id"]
                    .drop_duplicates()
                    .to_list()
                )
            LOGGER.info("depth %s: %s queries left", b, len(q_ids_left))

            # take the next chunk with b-a docs/passages for each query
            chunk = (
                df.loc[df["q_id"].isin(q_ids_left)]
                .groupby("q_id")
                .nth(list(range(a, b)))
            )

            # stop if no pairs are left
            if len(chunk) == 0:
                break

            # compute scores for the chunk and merge
            out = self._compute_scores(chunk, query_vectors)[["orig_index", "ff_score"]]
            chunk_scores = chunk.merge(out, on="orig_index", suffixes=(None, "_"))

            # compute interpolated scores
            chunk_scores["int_score"] = (
                alpha * chunk_scores["score"] + (1 - alpha) * chunk_scores["ff_score"]
            )

            scores_so_far = pd.concat(
                [scores_so_far, chunk_scores],
                axis=0,
            )

            a = b
        return scores_so_far.join(df, on="orig_index", lsuffix="", rsuffix="_")

    def __call__(
        self,
        ranking: Ranking,
        early_stopping: int | None = None,
        early_stopping_alpha: float | None = None,
        early_stopping_depths: Iterable[int] | None = None,
        batch_size: int | None = None,
    ) -> Ranking:
        """Compute scores for a ranking.

        :param ranking: The ranking to compute scores for. Must have queries attached.
        :param early_stopping: Perform early stopping at this cut-off depth.
        :param early_stopping_alpha: Interpolation parameter for early stopping.
        :param early_stopping_depths: Depths for early stopping.
        :param batch_size: How many queries to process at once.
        :raises ValueError: When the ranking has no queries attached.
        :raises ValueError: When early stopping is enabled but arguments are missing.
        :return: Ranking with the computed scores.
        """
        if not ranking.has_queries:
            raise ValueError("Input ranking has no queries attached.")
        if early_stopping is not None and (
            early_stopping_alpha is None or early_stopping_depths is None
        ):
            raise ValueError("Early stopping requires alpha and depths.")
        t0 = perf_counter()

        # get all unique queries and query IDs and map to unique numbers (0 to m)
        query_df = (
            ranking._df[["q_id", "query"]].drop_duplicates().reset_index(drop=True)
        )
        query_df["q_no"] = query_df.index
        df_with_q_no = ranking._df.merge(query_df, on="q_id", suffixes=(None, "_"))

        # early stopping splits the data frame, hence we need to keep track of the
        # original index
        df_with_q_no["orig_index"] = df_with_q_no.index

        # batch encode queries
        query_vectors = self.encode_queries(list(query_df["query"]))

        def _get_result(df: pd.DataFrame) -> pd.DataFrame:
            if early_stopping is None:
                return self._compute_scores(df, query_vectors)

            assert early_stopping_alpha is not None
            assert early_stopping_depths is not None
            return self._early_stopping(
                df,
                query_vectors,
                early_stopping,
                early_stopping_alpha,
                early_stopping_depths,
            )

        num_queries = len(query_df)
        if batch_size is None or batch_size >= num_queries:
            result = _get_result(df_with_q_no)
        else:
            # assign batch indices to query IDs
            df_with_q_no["batch_idx"] = (
                df_with_q_no.groupby("q_id").ngroup() / batch_size
            ).astype(int)

            chunks = []
            num_batches = int(num_queries / batch_size) + 1
            for batch_idx in tqdm(range(num_batches)):
                batch = df_with_q_no[df_with_q_no["batch_idx"] == batch_idx]
                chunks.append(_get_result(batch))
            result = pd.concat(chunks)

        result["score"] = result["ff_score"]
        LOGGER.info("computed scores in %s seconds", perf_counter() - t0)
        return Ranking(
            result,
            name="fast-forward",
            dtype=ranking._df.dtypes["score"],
            copy=False,
            is_sorted=False,
        )

    @abc.abstractmethod
    def _batch_iter(
        self, batch_size: int
    ) -> Iterator[tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over the index in batches.

        If a quantizer is used, the vectors are the quantized codes.
        When an ID does not exist, it must be set to `None`.

        Specific to index implementation.

        :param batch_size: The batch size.
        :return: Iterator yielding vectors, document IDs, passage IDs (in batches).
        """
        pass

    def batch_iter(
        self, batch_size: int
    ) -> Iterator[tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over all vectors, document IDs, and passage IDs in batches.

        IDs may be either strings or `None`.

        :param batch_size: Batch size.
        :return: Iterator yielding batches of vectors, document IDs (if any),
            passage IDs (if any).
        """
        if self._quantizer is None:
            yield from self._batch_iter(batch_size)

        else:
            for batch in self._batch_iter(batch_size):
                yield self._quantizer.decode(batch[0]), batch[1], batch[2]

    def __iter__(
        self,
    ) -> Iterator[tuple[np.ndarray, str | None, str | None]]:
        """Iterate over all vectors, document IDs, and passage IDs.

        :return: Iterator yielding vectors, document IDs (if any), passage IDs (if any).
        """
        for vectors, doc_ids, psg_ids in self.batch_iter(2**9):
            yield from zip(vectors, doc_ids, psg_ids)



import logging
from typing import TYPE_CHECKING, TypedDict

import numpy as np
from tqdm import tqdm

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from fast_forward.encoder.base import Encoder
    from fast_forward.index.base import IDSequence, Index
    from fast_forward.quantizer import Quantizer

LOGGER = logging.getLogger(__name__)


class IndexingDict(TypedDict):
    """Dictionary that represents a document/passage.

    Consumed by `Indexer.from_dicts`.
    """

    text: str
    doc_id: str | None
    psg_id: str | None


class Indexer:
    """Utility class for indexing collections."""

    def __init__(
        self,
        index: "Index",
        encoder: "Encoder | None" = None,
        encoder_batch_size: int = 128,
        batch_size: int = 2**16,
        quantizer: "Quantizer | None" = None,
        quantizer_fit_batches: int = 1,
    ) -> None:
        """Create an indexer.

        Optionally, a quantizer can automatically be fit on the first batch(es) to be
        indexed. This requires the index to be empty.
        If a quantizer is provided, the first batch(es) will be buffered and used to fit
        the quantizer.

        :param index: The target index.
        :param encoder: Document/passage encoder.
        :param encoder_batch_size: Encoder batch size.
        :param batch_size: How many vectors to add to the index at once.
        :param quantizer: A quantizer to be fit and attached to the index.
        :param quantizer_fit_batches: How many batches to use to fit the quantizer.
        :raises ValueError: When a quantizer is provided that has already been fit.
        :raises ValueError: When a quantizer is provided and the index is not empty.
        """
        self._index = index
        self._encoder = encoder
        self._encoder_batch_size = encoder_batch_size
        self._batch_size = batch_size
        self._quantizer = quantizer
        self._quantizer_fit_batches = quantizer_fit_batches

        if quantizer is not None:
            if quantizer._trained:
                raise ValueError(
                    "The quantizer is already fit. "
                    "It should be attached to the index directly."
                )

            if len(index) > 0:
                raise ValueError(
                    "The index must be empty for a quantizer to be attached."
                )

            self._buf_vectors, self._buf_doc_ids, self._buf_psg_ids = [], [], []
            if quantizer_fit_batches > 1:
                LOGGER.warning(
                    "inputs will be buffered and index will remain empty until the "
                    "quantizer has been fit"
                )

    def _index_batch(
        self,
        vectors: np.ndarray,
        doc_ids: "IDSequence | None" = None,
        psg_ids: "IDSequence | None" = None,
    ) -> None:
        """Add a batch to the index.

        If this indexer has a quantizer to be fit, the inputs will be buffered until the
        desired amount of data for fitting has been obtained. Afterwards, the quantizer
        is fit and attached to the index, and all buffered inputs are added at once.

        :param vectors: The vectors.
        :param doc_ids: Corresponding document IDs.
        :param psg_ids: Corresponding passage IDs.
        """
        if self._quantizer is None:
            self._index.add(vectors, doc_ids, psg_ids)
            return

        self._buf_vectors.append(vectors)
        self._buf_doc_ids.append(doc_ids)
        self._buf_psg_ids.append(psg_ids)

        if len(self._buf_vectors) >= self._quantizer_fit_batches:
            LOGGER.info(
                "fitting quantizer (%s batch(es), batch size %s)",
                len(self._buf_vectors),
                self._batch_size,
            )

            last_batch_size = self._buf_vectors[-1].shape[0]
            if last_batch_size < self._batch_size:
                LOGGER.warning(
                    "the size of the last batch (%s) is smaller than %s",
                    last_batch_size,
                    self._batch_size,
                )

            self._quantizer.fit(np.concatenate(self._buf_vectors))
            self._index.quantizer = self._quantizer
            self._quantizer = None

            LOGGER.info("adding buffered vectors to index")
            for b_vectors, b_doc_ids, b_psg_ids in zip(
                self._buf_vectors, self._buf_doc_ids, self._buf_psg_ids
            ):
                self._index.add(b_vectors, b_doc_ids, b_psg_ids)

            del self._buf_vectors
            del self._buf_doc_ids
            del self._buf_psg_ids

    def _encode(self, texts: "Sequence[str]") -> np.ndarray:
        """Encode a list of strings (respecting the encoder batch size).

        :param texts: The pieces of text to encode.
        :raises RuntimeError: When no encoder exists.
        :return: The vector representations.
        """
        if self._encoder is None:
            raise RuntimeError("An encoder is required.")

        result = []
        for i in range(0, len(texts), self._encoder_batch_size):
            batch = texts[i : i + self._encoder_batch_size]
            result.append(self._encoder(batch))
        return np.concatenate(result)

    def from_dicts(self, data: "Iterable[IndexingDict]") -> None:
        """Index data from dictionaries.

        :param data: An iterable of the dictionaries.
        """
        texts, doc_ids, psg_ids = [], [], []
        for d in tqdm(data):
            texts.append(d["text"])
            doc_ids.append(d.get("doc_id"))
            psg_ids.append(d.get("psg_id"))

            if len(texts) == self._batch_size:
                self._index_batch(self._encode(texts), doc_ids=doc_ids, psg_ids=psg_ids)
                texts, doc_ids, psg_ids = [], [], []

        if len(texts) > 0:
            self._index_batch(self._encode(texts), doc_ids=doc_ids, psg_ids=psg_ids)

    def from_index(self, index: "Index") -> None:
        """Transfer vectors and IDs from another index.

        If the source index uses quantized representations, the vectors are
        reconstructed first.

        :param index: The source index.
        """
        for vectors, doc_ids, psg_ids in tqdm(index.batch_iter(self._batch_size)):
            self._index_batch(vectors, doc_ids, psg_ids)



import logging
from collections import defaultdict
from typing import TYPE_CHECKING, cast

import h5py
import numpy as np
from tqdm import tqdm

import fast_forward
from fast_forward.index.base import IDSequence, Index, Mode
from fast_forward.index.memory import InMemoryIndex
from fast_forward.quantizer import Quantizer

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path

    from fast_forward.encoder.base import Encoder

LOGGER = logging.getLogger(__name__)


# h5py does not play nice with pyright, so we add lots of ignores in this class
class OnDiskIndex(Index):
    """Fast-Forward index that is read on-demand from disk.

    Uses HDF5 via h5py under the hood. The `max_indexing_size` argument works around an
    [h5py limitation](https://docs.h5py.org/en/latest/high/dataset.html#fancy-indexing).
    """

    def __init__(
        self,
        index_file: "Path",
        query_encoder: "Encoder | None" = None,
        quantizer: Quantizer | None = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        init_size: int = 2**14,
        resize_min_val: int = 2**10,
        hdf5_chunk_size: int | None = None,
        max_id_length: int = 8,
        overwrite: bool = False,
        max_indexing_size: int = 2**10,
    ) -> None:
        """Create an index on disk.

        :param index_file: The index file to create (or overwrite).
        :param query_encoder: The query encoder.
        :param quantizer: The quantizer to use.
        :param mode: The ranking mode.
        :param encoder_batch_size: Batch size for the query encoder.
        :param init_size: Initial size to allocate (number of vectors).
        :param resize_min_val: Minimum number of vectors to increase index size by.
        :param hdf5_chunk_size: Specify a chunk size used by HDF5.
        :param max_id_length:
            Maximum length of document and passage IDs (number of characters).
        :param overwrite: Overwrite index file if it exists.
        :param max_indexing_size:
            Maximum number of vectors to retrieve from the HDF5 dataset at once.
        :raises ValueError: When the file exists and `overwrite=False`.
        """
        if index_file.exists() and not overwrite:
            raise ValueError(f"File {index_file} exists.")

        self._index_file = index_file.absolute()
        self._doc_id_to_idx = defaultdict(list)
        self._psg_id_to_idx = {}

        self._init_size = init_size
        self._resize_min_val = resize_min_val
        self._hdf5_chunk_size = hdf5_chunk_size
        self._max_id_length = max_id_length
        self._max_indexing_size = max_indexing_size

        LOGGER.debug("creating file %s", self._index_file)
        with h5py.File(self._index_file, "w") as fp:
            fp.attrs["num_vectors"] = 0
            fp.attrs["ff_version"] = fast_forward.__version__

        super().__init__(
            query_encoder=query_encoder,
            quantizer=quantizer,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )

    def _on_quantizer_set(self) -> None:
        assert self.quantizer is not None

        # serialize the quantizer and store it on disk
        with h5py.File(self._index_file, "a") as fp:
            if "quantizer" in fp:
                del fp["quantizer"]

            meta, attributes, data = self.quantizer.serialize()
            fp.create_group("quantizer/meta").attrs.update(meta)
            fp.create_group("quantizer/attributes").attrs.update(attributes)
            data_group = fp.create_group("quantizer/data")
            for k, v in data.items():
                data_group.create_dataset(k, data=v)

    def _create_ds(self, fp: h5py.File, dim: int, dtype: np.dtype) -> None:
        """Create the HDF5 datasets for vectors and IDs.

        :param fp: Index file (write permissions).
        :param dim: Dimension of the vectors.
        :param dtype: Type of the vectors.
        """
        fp.create_dataset(
            "vectors",
            (self._init_size, dim),
            dtype,
            maxshape=(None, dim),
            chunks=True
            if self._hdf5_chunk_size is None
            else (self._hdf5_chunk_size, dim),
        )
        fp.create_dataset(
            "doc_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=True,
        )
        fp.create_dataset(
            "psg_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=True,
        )

    def _get_num_vectors(self) -> int:
        with h5py.File(self._index_file, "r") as fp:
            return fp.attrs["num_vectors"]  # pyright: ignore[reportReturnType]

    def _get_internal_dim(self) -> int | None:
        with h5py.File(self._index_file, "r") as fp:
            if "vectors" in fp:
                return fp["vectors"].shape[1]  # pyright: ignore[reportAttributeAccessIssue]
        return None

    def to_memory(self, batch_size: int | None = None) -> InMemoryIndex:
        """Load the index entirely into memory.

        :param batch_size: Use batches instead of adding all vectors at once.
        :return: The loaded index.
        """
        index = InMemoryIndex(
            query_encoder=self._query_encoder,
            quantizer=self._quantizer,
            mode=self.mode,
            encoder_batch_size=self._encoder_batch_size,
            init_size=len(self),
        )
        with h5py.File(self._index_file, "r") as fp:
            num_vectors = cast(int, fp.attrs["num_vectors"])

            batch_size = batch_size or num_vectors
            for i_low in range(0, num_vectors, batch_size):
                i_up = min(i_low + batch_size, num_vectors)

                doc_ids = fp["doc_ids"].asstr()[i_low:i_up]  # pyright: ignore[reportAttributeAccessIssue]
                psg_ids = fp["psg_ids"].asstr()[i_low:i_up]  # pyright: ignore[reportAttributeAccessIssue]
                vectors = fp["vectors"][i_low:i_up]  # pyright: ignore[reportIndexIssue]

                # IDs that don't exist will be returned as empty strings here
                doc_ids[doc_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                psg_ids[psg_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                index._add(vectors, doc_ids=doc_ids, psg_ids=psg_ids)  # pyright: ignore[reportArgumentType]
        return index

    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
    ) -> None:
        with h5py.File(self._index_file, "a") as fp:
            # if this is the first call to _add, no datasets exist
            if "vectors" not in fp:
                self._create_ds(fp, vectors.shape[-1], vectors.dtype)

            # check all IDs first before adding anything
            doc_id_size = fp["doc_ids"].dtype.itemsize  # pyright: ignore[reportAttributeAccessIssue]
            for doc_id in doc_ids:
                if doc_id is not None and len(doc_id) > doc_id_size:
                    raise RuntimeError(
                        f"Document ID {doc_id} is longer than the maximum "
                        f"({doc_id_size} characters)."
                    )
            psg_id_size = fp["psg_ids"].dtype.itemsize  # pyright: ignore[reportAttributeAccessIssue]
            for psg_id in psg_ids:
                if psg_id is not None and len(psg_id) > psg_id_size:
                    raise RuntimeError(
                        f"Passage ID {psg_id} is longer than the maximum "
                        f"({psg_id_size} characters)."
                    )

            num_new_vecs = vectors.shape[0]
            capacity = fp["vectors"].shape[0]  # pyright: ignore[reportAttributeAccessIssue]

            # check if we have enough space, resize if necessary
            cur_num_vectors = cast(int, fp.attrs["num_vectors"])
            space_left = capacity - cur_num_vectors
            if num_new_vecs > space_left:
                new_size = max(
                    cur_num_vectors + num_new_vecs,
                    capacity + self._resize_min_val,
                )
                LOGGER.debug("resizing index from %s to %s", capacity, new_size)
                for ds in ("vectors", "doc_ids", "psg_ids"):
                    fp[ds].resize(new_size, axis=0)  # pyright: ignore[reportAttributeAccessIssue]

            # add new document IDs to index and in-memory mappings
            doc_id_idxs, non_null_doc_ids = [], []
            for i, doc_id in enumerate(doc_ids):
                if doc_id is not None:
                    self._doc_id_to_idx[doc_id].append(cur_num_vectors + i)
                    doc_id_idxs.append(cur_num_vectors + i)
                    non_null_doc_ids.append(doc_id)
            fp["doc_ids"][doc_id_idxs] = non_null_doc_ids  # pyright: ignore[reportIndexIssue]

            # add new passage IDs to index and in-memory mappings
            psg_id_idxs, non_null_psg_ids = [], []
            for i, psg_id in enumerate(psg_ids):
                if psg_id is not None:
                    self._psg_id_to_idx[psg_id] = cur_num_vectors + i
                    psg_id_idxs.append(cur_num_vectors + i)
                    non_null_psg_ids.append(psg_id)
            fp["psg_ids"][psg_id_idxs] = non_null_psg_ids  # pyright: ignore[reportIndexIssue]

            # add new vectors
            fp["vectors"][cur_num_vectors : cur_num_vectors + num_new_vecs] = vectors  # pyright: ignore[reportIndexIssue]
            fp.attrs["num_vectors"] += num_new_vecs  # pyright: ignore[reportOperatorIssue]

    def _get_doc_ids(self) -> set[str]:
        return set(self._doc_id_to_idx.keys())

    def _get_psg_ids(self) -> set[str]:
        return set(self._psg_id_to_idx.keys())

    def _get_vectors(self, ids: "Iterable[str]") -> tuple[np.ndarray, list[list[int]]]:
        idx_pairs = []
        with h5py.File(self._index_file, "r") as fp:
            for id in ids:
                if self.mode in (Mode.MAXP, Mode.AVEP) and id in self._doc_id_to_idx:
                    idxs = self._doc_id_to_idx[id]
                elif self.mode == Mode.FIRSTP and id in self._doc_id_to_idx:
                    idxs = [self._doc_id_to_idx[id][0]]
                elif self.mode == Mode.PASSAGE and id in self._psg_id_to_idx:
                    idxs = [self._psg_id_to_idx[id]]
                else:
                    LOGGER.warning("no vectors for %s", id)
                    idxs = []

                for idx in idxs:
                    idx_pairs.append((id, idx))

            # h5py requires accessing the dataset with sorted indices
            idx_pairs.sort(key=lambda x: x[1])
            id_to_idxs = defaultdict(list)
            vec_idxs = []
            for id_idx, (id, vec_idx) in enumerate(idx_pairs):
                vec_idxs.append(vec_idx)
                id_to_idxs[id].append(id_idx)

            # reading all vectors at once slows h5py down significantly, so we read them
            # in chunks and concatenate
            vectors = np.concatenate(  # pyright: ignore[reportCallIssue]
                [
                    fp["vectors"][vec_idxs[i : i + self._max_indexing_size]]  # pyright: ignore[reportIndexIssue, reportArgumentType]
                    for i in range(0, len(vec_idxs), self._max_indexing_size)
                ]
            )
            return vectors, [id_to_idxs[id] for id in ids]

    def _batch_iter(
        self, batch_size: int
    ) -> "Iterator[tuple[np.ndarray, IDSequence, IDSequence]]":
        with h5py.File(self._index_file, "r") as fp:
            num_vectors = cast(int, fp.attrs["num_vectors"])
            for i in range(0, num_vectors, batch_size):
                j = min(i + batch_size, num_vectors)
                doc_ids = fp["doc_ids"].asstr()[i:j]  # pyright: ignore[reportAttributeAccessIssue]
                psg_ids = fp["psg_ids"].asstr()[i:j]  # pyright: ignore[reportAttributeAccessIssue]
                doc_ids[doc_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                psg_ids[psg_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                yield (
                    fp["vectors"][i:j],  # pyright: ignore[reportIndexIssue, reportReturnType]
                    doc_ids.tolist(),  # pyright: ignore[reportAttributeAccessIssue]
                    psg_ids.tolist(),  # pyright: ignore[reportAttributeAccessIssue]
                )

    @classmethod
    def load(
        cls,
        index_file: "Path",
        query_encoder: "Encoder | None" = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        resize_min_val: int = 2**10,
        max_indexing_size: int = 2**10,
    ) -> "OnDiskIndex":
        """Open an existing index on disk.

        :param index_file: The index file to open.
        :param query_encoder: The query encoder.
        :param mode: The ranking mode.
        :param encoder_batch_size: Batch size for the query encoder.
        :param resize_min_val: Minimum value to increase index size by.
        :param max_indexing_size:
            Maximum number of vectors to retrieve from the HDF5 dataset at once.
        :return: The index.
        """
        LOGGER.debug("reading file %s", index_file)

        index = cls.__new__(cls)
        super(OnDiskIndex, index).__init__(
            query_encoder=query_encoder,
            quantizer=None,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )
        index._index_file = index_file.absolute()
        index._resize_min_val = resize_min_val
        index._max_indexing_size = max_indexing_size

        # deserialize quantizer if any
        with h5py.File(index_file, "r") as fp:
            if "quantizer" in fp:
                index._quantizer = Quantizer.deserialize(
                    dict(fp["quantizer/meta"].attrs),  # pyright: ignore[reportArgumentType]
                    dict(fp["quantizer/attributes"].attrs),  # pyright: ignore[reportArgumentType]
                    {k: v[:] for k, v in fp["quantizer/data"].items()},  # pyright: ignore[reportAttributeAccessIssue]
                )

            # read ID mappings
            index._doc_id_to_idx = defaultdict(list)
            index._psg_id_to_idx = {}

            num_vectors = cast(int, fp.attrs["num_vectors"])
            if num_vectors == 0:
                return index

            doc_id_iter = fp["doc_ids"].asstr()[:num_vectors]  # pyright: ignore[reportAttributeAccessIssue]
            psg_id_iter = fp["psg_ids"].asstr()[:num_vectors]  # pyright: ignore[reportAttributeAccessIssue]
            for i, (doc_id, psg_id) in tqdm(
                enumerate(zip(doc_id_iter, psg_id_iter)),
                total=num_vectors,
            ):
                if len(doc_id) > 0:
                    index._doc_id_to_idx[doc_id].append(i)
                if len(psg_id) > 0:
                    index._psg_id_to_idx[psg_id] = i
        return index



""".. include:: ../docs/quantizer.md"""  # noqa: D400, D415

from fast_forward.quantizer.base import Quantizer
from fast_forward.quantizer.nanopq import NanoOPQ, NanoPQ

__all__ = ["Quantizer", "NanoPQ", "NanoOPQ"]
